# download sub-seasonal reforecast from S2S Database
from acacia_s2s_toolkit import argument_check, argument_output, webAPI_requests
import os
import sys
import datetime

class SuppressOutput:
    """Context manager to silence stdout/stderr (for ECMWF WebAPI logs)."""
    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = open(os.devnull, 'w')
        return self

    def __exit__(self, exc_type, exc, tb):
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = self._stdout
        sys.stderr = self._stderr
        # If an exception happened, return False so Python re-raises it (with logs visible next time)
        return False

def download_hindcast(model,
                         variable,
                         fcdate=None,
                         plevs=None,
                         location_name=None,
                         bbox_bounds=[90, -180, -90, 180],
                         filename=None,
                         data_save_dir=None,
                         data_format="netcdf",
                         grid="1.5/1.5",
                         leadtime_hour=None,
                         rf_years=None,
                         rf_enslags=None,
                         fc_time=True,
                         overwrite=False,
                         verbose=True):

    """
    Overarching function that will download hindcast data from ECDS.
    """

    # Domain bounds mapping
    DOMAIN_BOUNDS = {"ethiopia":   [15, 33, 3, 48],
                      "kenya":      [5, 33, -5, 42],
                      "madagascar": [-12, 43, -26, 51],
                      "eastafrica": [25, 10, -13, 55]}

    def match_domain_from_bbox_bounds(bbox_bounds):
        for name, bounds in DOMAIN_BOUNDS.items():
            if list(map(float, bounds)) == list(map(float, bbox_bounds)):
                return name
        return None

    def format_coord(value, lat=True):
        hemi = "N" if lat and value >= 0 else "S" if lat else "E" if value >= 0 else "W"
        return f"{abs(value)}{hemi}"

    # Domain overrides bbox_bounds
    if location_name is not None:
        cname =location_name.lower()
        if cname not in DOMAIN_BOUNDS:
            raise ValueError(f"Unsupportedlocation_name '{location_name}'. Choose from {list(DOMAIN_BOUNDS)}.")
        bbox_bounds = DOMAIN_BOUNDS[cname]
    else:
        cname = match_domain_from_bbox_bounds(bbox_bounds)
        if cname is not None:
           location_name = cname

    # Date handling with rollback to get the latest available forecast
    if fcdate is None:
        fcdate = datetime.datetime.utcnow().strftime("%Y%m%d")
  
    # output origin_id here
    model = model.upper() if any(c.islower() for c in model) else model
    origin_id = argument_output.output_originID(model,fcdate)

    while True:
        try:
            argument_check.check_fcdate(fcdate, origin_id)  # validate date
            break
        except ValueError:
            old_date = fcdate
            fcdate = (
                datetime.datetime.strptime(fcdate, "%Y%m%d")
                - datetime.timedelta(days=1)
            ).strftime("%Y%m%d")
            if verbose:
                print(f"[INFO] {old_date} not valid, rolling back to {fcdate}...")

    # Get parameters
    leveltype, plevs, webapi_param, ecds_varname, origin_id, leadtime_hour, fc_enslags = (argument_output.check_and_output_all_fc_arguments(variable, model, fcdate, bbox_bounds, data_format, grid, plevs, leadtime_hour, fc_enslags=0))

    print("[DEBUG] Download request details:")
    print(f"   Variable    : {variable}")
    print(f"   Domain     :  {location_name}")
    print(f"   BBox Bounds : {bbox_bounds}")
    print(f"   Forecast Date (fcdate): {fcdate}")
    print(f"   Pressure lvls: {plevs}")
    print(f"   Model       : {model}")
    print(f"   Origin ID   : {origin_id}")
    print(f"   Data format : {data_format}")
    print(f"   Grid        : {grid}")
    print(f"   Save Dir    : {data_save_dir}")
    print(f"   Filename    : {filename}")

    if rf_enslags == None:
        rf_enslags = argument_output.output_hc_lags(origin_id, fcdate)
    rf_model_date, rf_years = argument_output.check_and_output_all_hc_arguments(variable, origin_id, fcdate, rf_years)

    # Filename construction
    if filename is None:
        plev_str = ""
        if plevs is not None:
            plevs = [plevs] if isinstance(plevs, int) else plevs
            if len(plevs) == 1:
                plev_str = f"_{plevs[0]}hPa"
            else:
                plev_str = f"_{plevs[0]}-{plevs[-1]}hPa"

        if location_name:
            filename = f"{variable}_{model}_{fcdate}{plev_str}_{location_name}_hc"
        else:
            north, west, south, east = bbox_bounds
            bounds_str = (
                f"{format_coord(north, lat=True)}_{format_coord(west, lat=False)}_"
                f"{format_coord(south, lat=True)}_{format_coord(east, lat=False)}"
            )
            filename = f"{variable}_{model}_{fcdate}{plev_str}_{bounds_str}_hc"


    # Add extension
    filename_save = filename  # reassign var name as ther are intermediate file names saved
    ext = ".nc" if data_format.lower() == "netcdf" else ".grib"
    if not filename_save.endswith(ext):
        filename_save = f"{filename_save}{ext}"

    # Ensure save directory
    if data_save_dir is not None:
        os.makedirs(data_save_dir, exist_ok=True)
        filename_save = os.path.join(data_save_dir, os.path.basename(filename_save))

    # Skip download if file exists
    if os.path.exists(filename_save) and not overwrite:
        print(f"[INFO] File already exists: {filename_save}, skipping download.")
        return filename_save

    # Request download
    filename_save = filename_save[:-3] #drop the .nc extension

    # Print clean info
    if verbose:
        print(f"[INFO] Downloading {variable} (plevs={plevs}) for {location_name or bbox_bounds}")
        print(f"[INFO] Saving as {filename_save}")

    # Request with suppression logic
    # print(">>>> DEBUG PLEVS GOING TO REQUEST:", plevs)
    try:
        if verbose:
            webAPI_requests.request_hindcast(
                fcdate, origin_id, grid, variable, bbox_bounds, data_format, webapi_param,
                leadtime_hour, leveltype, filename_save, plevs, rf_enslags, rf_years, fc_time
            )
        else:
            with SuppressOutput():
                webAPI_requests.request_hindcast(
                    fcdate, origin_id, grid, variable, bbox_bounds, data_format, webapi_param,
                    leadtime_hour, leveltype, filename_save, plevs, rf_enslags, rf_years, fc_time
                )
    except Exception as e:
        # Re-raise but ensure logs aren't hidden if debugging
        print(f"[ERROR] Download failed for {filename_save}")
        raise

    return filename_save






# scripts relevant for merging multiple forecast
import xarray as xr
import numpy as np
import pandas as pd
import glob as glob
import os

def merge_all_ens_members(filename,leveltype):
    # open all ensemble members. drop step and time variables. Just use valid time.
    all_fcs = xr.open_mfdataset(f'{filename}_allens_*',engine='cfgrib',combine='nested',concat_dim='fc_init_member') # open mfdataset but have fc_init_member as a dimension
        
    if "fc_init_member" not in all_fcs.dims: # in case only one member is downloaded
        all_fcs = all_fcs.expand_dims("fc_init_member")

    if np.size(all_fcs.fc_init_member) == 1: # if only one forecast member is present
        if "valid_time" not in all_fcs.dims:
            if "valid_time" not in all_fcs.coords:
                all_fcs = all_fcs.expand_dims("valid_time")
            else:
                all_fcs = all_fcs.swap_dims({"step":"valid_time"})
        all_fcs = all_fcs.drop_vars(['step','time'])
    else: # concatenating forecasts from lagged ensemble. 
        os.system(f'rm {filename}*.idx') # annoying index files. these appear when reading the file in 'cfgrib'
        all_data = []
        # reopen files individually
        files = glob.glob(f'{filename}_allens_*')
        for file in files:
            single_fc = xr.load_dataset(file,engine='cfgrib')
            single_fc = single_fc.swap_dims({"step":"valid_time"})
            single_fc = single_fc.drop_vars(['step','time'])

            all_data.append(single_fc)
        all_fcs = xr.concat(all_data,dim='fc_init_member')

# OLD - ALL FCS! 
        #    # check for duplicate time points (common for accumulations)
        #    # SORT FOR LAGGED ENSEMBLE (NCEP for instance)
        #    if len(single_mem_sel['valid_time'].values) != len(pd.unique(single_mem_sel['valid_time'].values)):
        #        # There are duplicates → merge them along 'number'
        #        print ('Please note. There are still multiple values of the same time after merging valid_time (most likely due to an accumulation field that takes into consideration forecast step). To manually fix, we put all forecasts with the same timestamp in the same ensemble set. Further investigation needed to understand effect but for accumulation, you will most likely subtract from another time stamp')
        #        unique_times = pd.unique(single_mem_sel['valid_time'].values)

        #        slices = [single_mem_sel.sel(valid_time=t).stack(member=("number","valid_time")).reset_index("member", drop=True).assign_coords(valid_time=t) for t in unique_times]
        #        single_mem_sel = xr.concat(slices, dim='valid_time')
        #        single_mem_sel = single_mem_sel.rename({'member':'number'})

         #   all_data.append(single_mem_sel)
        #all_fcs = xr.concat(all_data,dim='fc_init_member')

    # stack, forecast init member and nunber
    combined = all_fcs.stack(member=("fc_init_member", "number")).reset_index("member", drop=True)
    combined = combined.rename({'valid_time':'time'})

    # put forecast initialisation time as an attribute
    try: # try adding forecast initialisation time (slightly off if lagged ensemble) - something to fix. 
        if np.size(all_fcs['time']) == 1:
            combined.attrs['Forecast_initialisation_time'] = str(all_fcs['time'].values)
    except:
        pass

    if leveltype == 'pressure':
        if 'isobaricInhPa' in combined.dims:
            combined = combined.rename({'isobaricInhPa':'level'})
        # Only transpose dims that actually exist
        combined = combined.transpose(
            *[d for d in ['time','member','level','latitude','longitude'] if d in combined.dims]
        )
    else:
        combined = combined.transpose(
            *[d for d in ['time','member','latitude','longitude'] if d in combined.dims]
        )

    return combined

def merge_all_ens_hindcasts(filename,leveltype):
    all_fcs = xr.open_mfdataset(f'{filename}_allens_*',combine='nested',concat_dim='fc_init_member') # open mfdataset but have fc_init_member as a dimension, i.e. number of forecast initialisations used.

    if "fc_init_member" not in all_fcs.dims:
        all_fcs = all_fcs.expand_dims("fc_init_member") # expand a fc_init_member if only one file is download. it will have a dimension of 1. 

    combined = all_fcs.stack(member=("fc_init_member", "number")).reset_index("member", drop=True)

    if leveltype == 'pressure':
        if 'isobaricInhPa' in combined.dims:
            combined = combined.rename({'isobaricInhPa':'level'})
        # Only transpose dims that actually exist
        combined = combined.transpose(
            *[d for d in ['time','member','level','latitude','longitude'] if d in combined.dims]
        )
    else:
        print (combined)
        combined = combined.transpose(
            *[d for d in ['time','member','latitude','longitude'] if d in combined.dims]
        )

    return combined




# Check that requested variables are appropriate and are compatiable with WMO lead centre.
from acacia_s2s_toolkit.variable_dict import s2s_variables
from difflib import get_close_matches
from datetime import datetime, timedelta
from acacia_s2s_toolkit import argument_output
import numpy as np

def check_requested_variable(variable):
    '''check requested variable matches abbreviations used for the S2S database.
    No return - ECDS variable outputted in variable_output.py
    '''
    
    # Flatten all variables from nested dictionary
    all_vars = []
    for category_dict in s2s_variables.values():
        for subcategory_vars in category_dict.values():
            all_vars.extend(subcategory_vars)

    if variable in all_vars:
        return True  # Variable is valid
    else:
        # Try suggesting closest matches
        suggestions = get_close_matches(variable, all_vars, n=3)
        suggestion_msg = f" Did you mean: {', '.join(suggestions)}?" if suggestions else ""
        
        raise ValueError(
            f"Invalid variable: '{variable}' is not in the S2S database.{suggestion_msg}"
        )

def check_model_name(model,fcdate):
    # Flatten all models from nested dictionary
    df = argument_output.read_lookup_table(fcdate)
    all_models = list(df['Model'].values)

    if model in all_models:
        return True  # Variable is valid
    else:
        # Try suggesting closest matches
        suggestions = get_close_matches(model, all_models, n=3)
        suggestion_msg = f" Did you mean: {', '.join(suggestions)}?" if suggestions else ""

        raise ValueError(
            f"Invalid model: '{model}' is not currently compatible with this toolbox. {suggestion_msg}"
            )

def check_fcdate(fcdate,origin_id):
    """
    Check if fcdate is a valid date string in the format 'YYYYMMDD'.
    Returns a datetime.date object if valid, or raises an error if invalid.
    """
    if not isinstance(fcdate, str):
        raise ValueError(f"[ERROR] Forecast date must be a string, got {type(fcdate)}.")
        return None

    try:
        date_obj = datetime.strptime(fcdate, '%Y%m%d')
    except ValueError:
        raise ValueError(f"[ERROR] '{fcdate}' is not in the correct format 'YYYYMMDD'.")
        return None

    # given originID check requested fcdate is permitted
    now = datetime.utcnow()

    # get time difference between requested time and now. 
    time_diff = now - date_obj # how many days bigger is now compared to requested time.

    # using origin name and origin_latency_hours dictionary, check time_diff is larger than numbers of hours.
    # get latency period
    min_numhours = argument_output.get_single_parameter(origin_id,fcdate,'Delay')

    # check min_numhours is smaller than time_diff. for instant, you cannot request an ECMWF forecast after 24 hours. 
    if time_diff < timedelta(hours=float(min_numhours)):
       raise ValueError(f"[ERROR] The time difference between now and requested forecast date {time_diff} is smaller than the required minimum amount of time for originID {origin_id}.") 

    # check that requested forecast date, matches avaliable forecast initilisations
    # get weekday forecast initialisations
    weekdays_aval = argument_output.get_single_parameter(origin_id,fcdate,'fcFreq')

    # check that the forecast date weekday is avaliable for that model
    fcdate_weekday = date_obj.weekday()+1 # Monday = 1 etc..
    if fcdate_weekday not in weekdays_aval:
        raise ValueError(f"[ERROR] The chosen forecast initialisation date is not avaliable for the chosen model. Origin ID: {origin_id}")

def check_dataformat(data_format):
    if data_format not in ['grib','netcdf']:
        raise ValueError(f"[ERROR] The chosen data format is not avaliable. Please use 'grib' or 'netcdf'")

def check_leadtime_hours(leadtime_hour,variable,origin_id,fcdate):
    # is the maximum lead time smaller or equal to forecat end time
    end_time = argument_output.get_single_parameter(origin_id,fcdate,'fcLength')

    if np.max(leadtime_hour) > end_time:
        raise ValueError(f"[ERROR] You are requesting a leadtime greater than end of forecast, {end_time} hours")

    if np.min(leadtime_hour) < 0:
        raise ValueError(f"[ERROR] You are requesting a negative leadtime")

    # the check depends on time resolution
    time_resolution = argument_output.get_timeresolution(variable)

    if time_resolution.endswith('6hrly'):
        output_freq = 6
    else:
        output_freq = 24

    if not np.all(leadtime_hour % output_freq == 0):
        raise ValueError(f"[ERROR] You are requesting a leadtime that is not compatible with output frequency. Output frequency of the desired variable is {output_freq} hours. You are requesting the following {leadtime_hour}.")

def check_plevs(plevs,variable):
    # first get maximum plevs avaliable
    max_plevs = argument_output.output_plevs(variable)

    # Check all requested plevs are valid
    if not np.all(np.isin(plevs, max_plevs)):
        invalid = [p for p in plevs if p not in max_plevs]
        raise ValueError(
            f"[ERROR] Invalid pressure level(s) requested: {invalid}. "
            f"Available levels for '{variable}' are: {max_plevs}."
        )

def check_area_selection(area):
    # Go through each component of the area (N, W, S, E)
    if area[0] < -90 or area[0] > 90:
        raise ValueError(
            f"[ERROR] Invalid northern latitude '{area[0]}'. Must be between -90 and 90."
            )

    if area[1] < -180 or area[1] > 180:
        raise ValueError(
            f"[ERROR] Invalid western longitude '{area[1]}'. Must be between -180 and 180."
            )

    if area[2] < -90 or area[2] > 90:
        raise ValueError(
            f"[ERROR] Invalid southern latitude '{area[2]}'. Must be between -90 and 90."
            )

    if area[3] < -180 or area[3] > 180:
        raise ValueError(
           f"[ERROR] Invalid eastern longitude '{area[3]}'. Must be between -180 and 180."
        )

    # is north > south?
    if area[0] < area[2]:
        raise ValueError(
                f"[ERROR] Northern latitude {area[0]} must be greater than southern latitude {area[2]}."
            )
    if area[1] > area[3]:
         raise ValueError(
                f"[ERROR] Western longitude {area[1]} must be smaller than eastern longitude {area[3]}."
            )

def check_fc_enslags(fc_enslags):
    """
    Check that all values in fc_enslags are non-positive integers (i.e., ≤ 0 and whole numbers).
    Raises ValueError if check fails.
    """

    # Case 1: a single integer
    if isinstance(fc_enslags, int):
        if fc_enslags > 0:
            raise ValueError("All ensemble lags (fc_enslags) must be integers ≤ 0 (e.g., [0, -1, -2]).")

    # Case 2: iterable of integers, a list
    else:
        if not all(isinstance(lag, int) and lag <= 0 for lag in fc_enslags):
            raise ValueError("All ensemble lags (fc_enslags) must be integers ≤ 0 (e.g., [0, -1, -2]).")

def check_requested_reforecast_years(rf_years,origin_id,fc_date):
    ''' 
    Check that the requested reforecast years are able to download
    '''
    # first get full set of reforecast years
    full_rf_years = argument_output.get_hindcast_year_span(origin_id,fc_date)
   
    # check all years in rf_years are in full_rf_years
    if not all(year in full_rf_years for year in rf_years):
        raise ValueError(f"All requested reforecast years {rf_years} are not avaliable. Avaliable years are {full_rf_years}.")



# script containing all relevant code for producing webAPI requests.
# In Q3 2025, will move from webAPI to ECDS

from acacia_s2s_toolkit import argument_output, merge_lag_ensemble
import numpy as np
import os
import eccodes as ec
import xarray as xr
from datetime import datetime, timedelta
from ecmwfapi import ECMWFDataServer
server = ECMWFDataServer()

def create_initial_webAPI_request(fcdate,grid,area,origin,webapi_param,leadtimes,filename):
    request_dict = {
            "dataset": "s2s",
            "class": "s2",
            "date": f"{fcdate}",
            "expver": "prod",
            "grid": f"{grid}",
            "levtype": "sfc",
            "origin": f"{origin}",
            "param": f"{webapi_param}",
            "step": f"{leadtimes}",
            "time": "00:00:00",
            "stream": "enfo",
            "type": "cf",
            "target": f"{filename}"
            }

    # add area component
    area_formatted = '/'.join(str(x) for x in area)
    request_dict['area'] = f"{area_formatted}"

    return request_dict

def request_forecast(fcdate,origin,grid,variable,area,data_format,webapi_param,leadtime_hour,leveltype,filename,plevs,fc_enslags):
    # to enable lagged ensemble, loop through requested ensembles
    for lag in np.atleast_1d(fc_enslags):
        lag = int(lag)
        leadtimes, convert_fcdate = argument_output.output_formatted_leadtimes(leadtime_hour,fcdate,variable,origin,lag=lag,fc_enslags=fc_enslags) # you need to know new fcdate, but also appropriate leadtimes so all are forecasting same period
        # create initial control request
        request_dict = create_initial_webAPI_request(convert_fcdate,grid,area,origin,webapi_param,leadtimes,f'{filename}_control_{lag}')

        # change components of request based on level type, and grid
        # if grid doesn't equal '1.5/1.5', add 'repres' dictionary item which sets the requested representation, in this case, 'll'=latitude/longitude.
        if grid != '1.5/1.5':
            # add repres
            request_dict['repres'] = 'll'

        # if a pressure level type is selected, just need to change levtype and add list of pressure levels.
        if leveltype == 'pressure':
            request_dict['levtype'] = 'pl'
            # convert plevs
            plevels = '/'.join(str(x) for x in plevs)
            request_dict['levelist'] = f"{plevels}"

        # specific change needed for pv
        if variable == 'pv':
            request_dict['levtype'] = 'pt'
            request_dict['levelist'] = '320'

        # retrieve the control forecast
        server.retrieve(request_dict)

        # then download perturbed. change type of forecast, add number of ensemble members, and change target filename
        request_dict['type'] = 'pf'
        # add model number (will not be needed for ECDSapi)
        num_pert_fcs = argument_output.get_single_parameter(origin,fcdate,'fcNumEns')
        pert_fcs = '/'.join(str(x) for x in np.arange(1,num_pert_fcs+1))
        request_dict['number'] = f"{pert_fcs}"
        request_dict['target'] = f"{filename}_perturbed_{lag}"

        server.retrieve(request_dict)

        # once requesting control and perturbed forecast, combine the two.
        # set forecast type in control to pf (perturbed forecast).
        set_cf_to_pf(f'{filename}_control_{lag}',f'{filename}_control2_{lag}')
        # merge both control and perturbed forecast
        os.system(f'cdo merge {filename}_control2_{lag} {filename}_perturbed_{lag} {filename}_allens_{lag}')

    # create new 'member' dimension based on same date. For instance, 5 members per date and three initialisations used
    # smae process following even with one forecast initialisation date to ensure same structure for all output. 
    combined_forecast = merge_lag_ensemble.merge_all_ens_members(f'{filename}',leveltype)
    combined_forecast.to_netcdf(f'{filename}.nc')

    # remove previous files  
    os.system(f'rm {filename}_control* {filename}_perturbed* {filename}_allens*')

def request_hindcast(fcdate,origin,grid,variable,area,data_format,webapi_param,leadtime_hour,leveltype,filename,plevs,rf_enslags,rf_years,fc_time=True):
    # to enable lagged ensemble, loop through requested ensembles
    print (rf_enslags)
    for lag in np.atleast_1d(rf_enslags):
        lag = int(lag)
        # convert fcdate
        lagged_fcdate = datetime.strptime(fcdate, '%Y%m%d')+timedelta(days=lag) # work out what the lagged fcdate is.
        convert_fcdate = lagged_fcdate.strftime('%Y%m%d') # convert that date to YYYYMMDD format

        rf_model_date, rfyears = argument_output.check_and_output_all_hc_arguments(variable,origin,convert_fcdate,rf_years) # get the reforecast model date version plus the number of reforecast years

        leadtimes, lag_fcdate_dashed_format = argument_output.output_formatted_leadtimes(leadtime_hour,convert_fcdate,variable,origin) # get the appropriate leadtimes plus the convert fcdate.
        print (leadtimes)

        # create initial control request
        request_dict = create_initial_webAPI_request(lag_fcdate_dashed_format,grid,area,origin,webapi_param,leadtimes,f'{filename}_control_{lag}')

        # use correct reforecast model date
        request_dict['date'] = f"{rf_model_date}"

        # download reforecast, so change stream
        request_dict['stream'] = f"enfh"

        # create list of hdates
        hdates = argument_output.create_reforecast_dates(rfyears,convert_fcdate)
        request_dict['hdate']=f"{hdates}"

        # change components of request based on level type, and grid
        # if grid doesn't equal '1.5/1.5', add 'repres' dictionary item which sets the requested representation, in this case, 'll'=latitude/longitude.
        if grid != '1.5/1.5':
            # add repres
            request_dict['repres'] = 'll'

        # if a pressure level type is selected, just need to change levtype and add list of pressure levels.
        if leveltype == 'pressure':
            request_dict['levtype'] = 'pl'
            # convert plevs
            plevels = '/'.join(str(x) for x in plevs)
            request_dict['levelist'] = f"{plevels}"

        # specific change needed for pv
        if variable == 'pv':
            request_dict['levtype'] = 'pt'
            request_dict['levelist'] = '320'

        # retrieve the control forecast
        server.retrieve(request_dict)

        # then download perturbed. change type of forecast, add number of ensemble members, and change target filename
        request_dict['type'] = 'pf'
        # add model number (will not be needed for ECDSapi)
        num_pert_hcs = argument_output.get_single_parameter(origin,convert_fcdate,'rfNumEns')
        pert_hcs = '/'.join(str(x) for x in np.arange(1,num_pert_hcs+1))
        request_dict['number'] = f"{pert_hcs}"
        request_dict['target'] = f"{filename}_perturbed_{lag}"

        server.retrieve(request_dict)

        # once requesting control and perturbed forecast, combine the two.
        # set forecast type in control to pf (perturbed forecast).
        set_cf_to_pf(f'{filename}_control_{lag}',f'{filename}_control2_{lag}')
        # merge both control and perturbed forecast
        os.system(f'cdo merge {filename}_control2_{lag} {filename}_perturbed_{lag} {filename}_allens_{lag} 2>/dev/null')
        # shift the time so all reforecasts have the same time values
        if fc_time:
            shift_day_value = lag*-1
        else:
            shift_day_value = 0 
        # only shift the time, if you want a 'forecast-based' time.
        rf_shifttime(f'{filename}_allens_{lag}',f'{filename}_timeshifted_allens_{lag}',shift_days=shift_day_value)

    # create new 'member' dimension based on same date. For instance, 5 members per date and three initialisations used
    # same process following even with one forecast initialisation date to ensure same structure for all output. 
    combined_forecast = merge_lag_ensemble.merge_all_ens_hindcasts(f'{filename}_timeshifted',leveltype)
    combined_forecast.to_netcdf(f'{filename}.nc')

    # remove previous files  
    os.system(f'rm {filename}_control* {filename}_perturbed* {filename}_*allens*')

def rf_shifttime(fn,output_fn,shift_days=0):
    orig_hc = xr.open_dataset(fn,engine='cfgrib')
    if np.size(orig_hc['step'].values) > 1: # different option depending on size of step variable.
        data = [] # loop through each time point (forecast initialisation). only save the valid forecast time.
        for time_pt in orig_hc['time']:
            select = orig_hc.sel(time=time_pt)
            select = select.assign_coords(step=select['valid_time'])
            select = select.drop_vars(["valid_time","time"])
            select = select.rename({'step':'valid_time'})
            data.append(select)
        orig_hc_new = xr.merge(data)
    else:
        orig_hc_new = orig_hc.swap_dims({'time':'valid_time'})
        orig_hc_new = orig_hc_new.drop_vars("time")

    orig_hc_new = orig_hc_new.assign_coords(
            valid_time=orig_hc_new.valid_time+np.timedelta64(shift_days,'D')
            )
    orig_hc_new = orig_hc_new.rename({'valid_time':'time'})
    orig_hc_new.to_netcdf(output_fn)

def set_cf_to_pf(input_file, output_file):
    # Open input file
    with open(input_file, 'rb') as fin,open(output_file,'wb') as fout:
        while True:
            gid = ec.codes_grib_new_from_file(fin)
            if gid is None:
                break
            # Set type from 'cf' to 'pf'
            ec.codes_set(gid, 'type', 'pf')
            ec.codes_write(gid,fout)
            ec.codes_release(gid)




# output suitable ECDS variables in light of requested forecasts.
from acacia_s2s_toolkit import variable_dict, argument_check
from datetime import datetime, timedelta
from importlib import resources
import numpy as np
import pandas as pd
import ast

def read_lookup_table(fcdate='20250828'):
    # convert fcdate to a time field. 
    fc_dt = datetime.strptime(fcdate,'%Y%m%d')
    csv_dir = "acacia_s2s_toolkit.lookup_tables"
   
    first_lookup_table_date = '20240514'
    first_lookup_table_time = datetime.strptime(first_lookup_table_date,'%Y%m%d')

    ############ new lookup table directory ##### Select the most recent lookup table compared to the requested forecast date #####
    if fc_dt < first_lookup_table_time:
        csv_file = f'lookup_table_{first_lookup_table_date}.csv'
    else:
        # list all files in directory
        files = [f.name for f in resources.files(csv_dir).iterdir() if f.name.startswith("lookup_table_") and f.name.endswith(".csv")]

        file_dates = [(datetime.strptime(f.split("_")[-1].replace(".csv",""),"%Y%m%d"),f) for f in files] # outputs dates + name of files.

        # select valid dates (all less than fcdate)
        valid_dates = [(date,file) for date, file in file_dates if date <= fc_dt]
        if not valid_dates:
            raise FileNotFoundError(f"No lookup table found before {fcdate}")

        # pick the most recent one, then read that lookup table
        chosen_date, csv_file = max(valid_dates, key=lambda x:x[0])

    with resources.open_text(csv_dir,csv_file) as fc_info:
        df = pd.read_csv(fc_info)

    columns_to_eval = ["fcFreq", "dayfcLags", "rfRange", "rfLagDetail"]

    for col in columns_to_eval:
        # if the option is a list of numbers or zero.
        df[col] = df[col].apply(lambda x: int(x) if x == "0" else (ast.literal_eval(x) if isinstance(x, str) and (x.startswith("[") or x.startswith("(")) else x)) 
    return df

def get_single_parameter(origin_id,fcdate,parameter):
    # first read lookup table
    df = read_lookup_table(fcdate)
    # find forecast length for originID
    match = df.loc[df["Origin"] == origin_id, parameter]

    if match.empty:
        print (f"[ERROR] could not find {parameter} for originID '{origin_id}'.")
        return None

    return match.iloc[0]

def get_timeresolution(variable):
    # first find which sub-category the variable sits in
    time_resolution=None
    for category_name, category_dict in variable_dict.s2s_variables.items():
        for subcategory_name, subcategory_vars in category_dict.items():
            if variable in subcategory_vars:
                time_resolution = subcategory_name
                break # found correct time resolution
        if time_resolution:
            break # break outer loop

    if time_resolution is None:
        print (f"[ERROR] could not find variable '{variable}'.")
        return None
    return time_resolution

def output_leadtime_hour(variable,origin_id,fcdate,fc_enslags,start_time=0):
    '''
    Given variable (variable abbreivation), output suitable leadtime_hour. The leadtime_hour will request all avaliable steps. Users should be able to pre-define leadtime_hour if they do not want all output.
    return: leadtime_hour
    '''
    time_resolution = get_timeresolution(variable)

    # next find maximum end time
    end_time = get_single_parameter(origin_id,fcdate,'fcLength')

    # given time resolution, work out array of appropriate time values
    if time_resolution.endswith('6hrly'):
        leadtime_hour = np.arange(start_time,end_time+1,6)
    else:
        leadtime_hour = np.arange(start_time,end_time+1,24) # will output 0 to 1104 in steps of 24 (ECMWF example). 
 
    # take the minimum value in fc_enslags and multiply by 24
    #max_endtime = end_time+(24*np.min(fc_enslags)) # so 1104 + (24*-2) = 1056
    #leadtime_hour = leadtime_hour[leadtime_hour <= max_endtime]

    return leadtime_hour

def output_sfc_or_plev(variable):
    '''
    Given variable (variable abbreivation), output whether variable is sfc level or on pressure levels?
    return: level_type
    '''
    # Flatten all variables from nested dictionary
    level_type=None
    for category_name, category_dict in variable_dict.s2s_variables.items():
        for subcategory_vars in category_dict.values():
            if variable in subcategory_vars:
                level_type = category_name
                return level_type
    if level_type == None:
        print (f"[ERROR] No leveltype found for '{variable}'.")
        return level_type

def output_webapi_variable_name(variable):
    ''' 
    Given variable abbreviation, output webAPI paramID.
    return webAPI paramID.

    '''
    for variable_abb, webapi_code in variable_dict.webAPI_params.items():
        if variable == variable_abb:
            return webapi_code
    print (f"[ERROR] No webAPI paramID found for '{variable}'.")
    return None

def output_originID(model,fcdate):
    '''
    Given model name, output originID.
    return originID.

    '''
    # first read lookup table
    df = read_lookup_table(fcdate)
    # find forecast length for originID
    match = df.loc[df["Model"] == model, "Origin"]

    if match.empty:
        print (f"[ERROR] could not find for originID for model '{model}'.")
        return None

    return match.iloc[0]

def output_ECDS_variable_name(variable):
    '''
    Given variable name, output the matching ECDS variable name
    
    return ECDS_varname (ECMWF Data Store)
    '''
    ECDS_varname='10m_uwind'
    return ECDS_varname

def output_plevs(variable):
    '''
    Output suitable plevs, if q, (1000, 925, 850, 700, 500, 300, 200) else add 100, 50 and 10 hPa. 
    '''
    all_plevs=[1000,925,850,700,500,300,200,100,50,10]
    if variable == 'q':
        plevs=all_plevs[:-3] # if q is chosen, don't download stratosphere
    else:
        plevs=all_plevs
    print (f"Selected the following pressure levels: {plevs}")
    
    return plevs

def output_fc_lags(origin_id,fcdate):
    '''
    Given origin_id, output lagged ensemble forecasts.
    return array with day lag positions, i.e. [0,-1,-2].
    '''
    fclags = get_single_parameter(origin_id,fcdate,'dayfcLags')

    # Special handling for CPTEC (sbsj). Initialisation only given for Wednesday and Thursday.
    if origin_id == 'sbsj':
        date_obj = datetime.strptime(fcdate, '%Y%m%d')
        weekday = date_obj.weekday()+1  # Monday = 1, ..., Sunday = 7
        if weekday == 4:  # Thursday
            return [0, -1]
        else:
            return [0]
    # Return the list from the dictionary
    return fclags

def output_hc_lags(origin_id,fcdate):
    '''
    Given origin_id, outputted the best lags for downloading reforecasts
    '''
    lag_type = get_single_parameter(origin_id,fcdate,'rfLagType')
    rf_freq_info = get_single_parameter(origin_id,fcdate,'rfLagDetail')

    fcdate_obj = datetime.strptime(fcdate, '%Y%m%d')
    weekday = fcdate_obj.weekday()+1  # Monday = 1, ..., Sunday = 7
    dayofmonth = fcdate_obj.day

    # ---- Nearest DOM (single closest) -------
    if "nearestDOM" == lag_type: # DOMs. get closest lags
        # Build candidate reforecast dates in prev/current/next month
        candidates = []
        for offset_month in [-1, 0, 1]:
            # Month/year shift
            month = (fcdate_obj.month - 1 + offset_month) % 12 + 1
            year = fcdate_obj.year + ((fcdate_obj.month - 1 + offset_month) // 12)
            for dom in rf_freq_info:
                try:
                    candidates.append(datetime(year, month, dom))
                except ValueError:
                    continue  # skip invalid dates (e.g., Feb 30)

        # Find the one with smallest abs(lag)
        closest_rf = min(candidates, key=lambda d: abs((d - fcdate_obj).days))
        lag = (closest_rf - fcdate_obj).days
        return lag
    # ------ Before/after DOM -----------------
    if "before_after_DOM" == lag_type:
        # Build candidate reforecast dates in prev/current/next month
        candidates = []
        for offset_month in [-1, 0, 1]:
            # Month/year shift
            month = (fcdate_obj.month - 1 + offset_month) % 12 + 1
            year = fcdate_obj.year + ((fcdate_obj.month - 1 + offset_month) // 12)
            for dom in rf_freq_info:
                try:
                    candidates.append(datetime(year, month, dom))
                except ValueError:
                    continue  # skip invalid dates (e.g., Feb 30)

        lags = [(c-fcdate_obj).days for c in candidates]

        neg_lags = [l for l in lags if l < 0]
        pos_lags = [l for l in lags if l >= 0]

        largest_neg = max(neg_lags) if neg_lags else None
        smallest_pos = min(pos_lags) if pos_lags else None

        return [largest_neg,smallest_pos] # before and after lags
    # ------ daily reforecast initialisations ------
    if "daily_lagged" == lag_type:
        return rf_freq_info

    # ------ Weekday based (i.e. only Monday and Thursday reforecasts)-----
    if "weekday_based" == lag_type:
        # two options depending on whether fcdate is a Monday or Thursday. 
        if fcdate_obj >= datetime(2023,6,27) and origin_id == 'ecmf': # for ecmwf between 20230627 and 20241112, ECMWF ran daily forecasts but MoThMo reforecasts. 
            return two_Monday_Thursday_rfdates(fcdate_obj)
        else:
            if weekday == 4: # Thursday
                return [-3,0]
            elif weekday == 1: # Monday
                return [0,3]
            else:
                raise ValueError(f"[ERROR] For origin_id '{origin_id}' forecasts are only every Monday and Thursday, therefore a Monday or Thursday forecast date must be selected.")

    # ---- Unique modes  ------
    if "unique" == lag_type:
        # ECMWF ---- odd day reforecasts.
        if rf_freq_info == 'odddates':
            if (dayofmonth % 2 == 0) or (dayofmonth == 29 and fcdate_obj.month() == 2):
                return [-1,1] # for even fcdates (or 29th Feb) chosen rfdate before and after forecast date.
            else:
                return [-2,0] # for odd dates, choosen current day, minus 2. 
        if rf_freq_info == 'CNRevery5days': # roughly five days
            # make an array of dates from 2020-01-01 to 2020-12-27
            DOM = {1:[1,6,11,16,21,26,31],2:[5,10,15,20,25],3:[2,7,12,17,22,27],4:[1,6,11,16,21,26],5:[1,6,11,16,21,26,31],6:[5,10,15,20,25,30],7:[5,10,15,20,25,30],8:[4,9,14,19,24,29],9:[3,8,13,18,23,28],10:[3,8,13,18,23,28],11:[2,7,12,17,22,27],12:[2,7,12,17,22,27]}
            CNR_rf_dates = []
            for year in (2020,2021):
                for month, days in DOM.items():
                    for day in days:
                        CNR_rf_dates.append(datetime(year,month,day))
            # change year of fcdate to 2020-fcdate(MM)-fcdate(DD)
            fc_date_2020 = datetime.strptime(f"2020{fcdate[4:]}",'%Y%m%d')
            # nearest date to altered fcdate is rf_date
            closest_day=min(CNR_rf_dates,key=lambda x:abs(fc_date_2020-x))
            lag = (closest_day - fc_date_2020).days
            return lag
        if rf_freq_info == 'JMAtwicepermonth': # twice per month
            # make an array of dates from 2020-01-01 to 2020-12-27
            DOM = {1:[16,31],2:[10,25],3:[12,27],4:[11,26],5:[16,31],6:[15,30],7:[15,30],8:[14,29],9:[13,28],10:[13,28],11:[12,27],12:[12,27]}
            JMA_rf_dates = []
            for year in (2019,2020,2021):
                for month, days in DOM.items():
                    for day in days:
                        JMA_rf_dates.append(datetime(year,month,day))
            # change year of fcdate to 2020-fcdate(MM)-fcdate(DD)
            fc_date_2020 = datetime.strptime(f"2020{fcdate[4:]}",'%Y%m%d')

            lags = [(c-fc_date_2020).days for c in JMA_rf_dates]
            neg_lags = [l for l in lags if l < 0]
            pos_lags = [l for l in lags if l >= 0]

            largest_neg = max(neg_lags) if neg_lags else None
            smallest_pos = min(pos_lags) if pos_lags else None

            return [largest_neg,smallest_pos] 

def two_Monday_Thursday_rfdates(fcdate):
    '''
    Given a fcdate, output the appropriate three reforecast start dates. Relevant for ECMWF operations between 27th June 2023 and 12th November 2024
    '''
    # create a list of dates +- 14 days around fcdate
    pos_dates = [fcdate + timedelta(days=delta) for delta in range(-14,15)]
    # keep only Mondays and Thursdays
    mon_thurs = [date for date in pos_dates if date.weekday() in (0,3)]
    days_delta = [(md-fcdate).days for md in mon_thurs]
    # then sort by abs value
    days_delta.sort(key=lambda x: (abs(x)))

    return days_delta[:2]


def get_hindcast_model_date(origin_id,fcdate):
    ''' Given origin_id, output appropriate date for reforecast dataset. This is the hindcast model version, not the set of reforecast dates.
    '''
    rf_model_freq = get_single_parameter(origin_id,fcdate,'rfModelFreq')
    rf_model_date = get_single_parameter(origin_id,fcdate,'rfModelDate')
    
    if rf_model_freq == "fixed":
        mrf_date = str(rf_model_date)
    else:
        mrf_date = fcdate

    return mrf_date
        
def get_hindcast_year_span(origin_id,fcdate):
    ''' Given origin_id, output appropriate set of years for reforecasts.
    '''
    rfType = get_single_parameter(origin_id,fcdate,'rfType')
    rf_yrs = get_single_parameter(origin_id,fcdate,'rfRange')

    if rfType == "fixed":
        start, end = rf_yrs
        rf_years = np.arange(start,end+1) # give full set of years, i.e. 1981, 1982, ..., 2013.
    elif rfType == "dynamic":
        fc_year = int(fcdate[:4]) # get year component of date.
        rf_years = np.arange(fc_year - int(rf_yrs),fc_year)
    else:
        raise ValueError(f"[ERROR] Couldn't compute appropriate reforecast years for origin_id '{origin_id}'.")

    return rf_years

def output_formatted_leadtimes(leadtime_hour,fcdate,variable,origin_id,lag=0,fc_enslags=0):
    ''' lag is always negative for forecast. Function always uses lag=0 for reforecast download'''

    print (fc_enslags)

    # create new fcdate based on lag
    new_fcdate = datetime.strptime(fcdate, '%Y%m%d')+timedelta(days=float(lag))
    convert_fcdate = new_fcdate.strftime('%Y-%m-%d')

    # check last leadtime is small enough to handle
    # first get fcLength
    fc_length = get_single_parameter(origin_id,fcdate,'fcLength')
    if np.max(leadtime_hour) > fc_length-(24.0*np.min(fc_enslags)):
        raise ValueError(f"[ERROR] The maximum requested leadtime hour is greater than the forecast length - the largest change in lagged ensemble (hours).")

    # convert leadtimes
    # is it an average field?
    time_resolution = get_timeresolution(variable)

    # get standard, all leadtime hours avaliable
    leadtime_hour_ALL = output_leadtime_hour(variable,origin_id,fcdate,fc_enslags)

    # find the indexes of the requested leadtimes (leadtime_hour)
    idx = np.asarray([np.where(leadtime_hour_ALL == lt)[0][0] for lt in leadtime_hour]) # get the indexes of where the requested leadtime hours are

    # if an average field, use '0-24/24-48/48-72...'
    leadtime_hour_copy = leadtime_hour[:]

    if time_resolution.endswith('6hrly'):
        nsteps_per_day = 4
        new_idx = idx-lag*nsteps_per_day
    else: # instantaneous field
        new_idx = idx-lag

    # check index is no larger than forecast length
    if np.max(new_idx) > len(leadtime_hour_ALL):
        raise ValueError(f"[ERROR] The maximum requested leadtime hour is greater than the forecast length, i.e. when performing lagged forecast ensemble, it can reach the last few timesteps.")

    leadtime_hour_copy=leadtime_hour_ALL[new_idx]

    if time_resolution.startswith('aver'):
        leadtimes='/'.join(f"{leadtime_hour_copy[i]}-{leadtime_hour_copy[i]+24}" for i in range(len(leadtime_hour_copy)-1))
    else: # instantaneous field
        leadtimes = '/'.join(str(x) for x in leadtime_hour_copy)
    
    print (leadtimes)
    
    return leadtimes, convert_fcdate

def create_reforecast_dates(rfyears,rfdate):
    ''' function that produces a list of reforecast dates given set of years and chosen reforecast date
    '''
    if np.size(rfdate) == 1: # for a single reforecast date that is then repeated for all reforecast years
        DOY = rfdate[4:]
        rf_dates = '/'.join(f"{int(year)}{DOY}" for year in rfyears)
    else:
        rf_dates = '/'.join(f"{date}" for date in rfdate) 
    return rf_dates
 

def check_and_output_all_fc_arguments(variable,model,fcdate,area,data_format,grid,plevs,leadtime_hour,fc_enslags):
    # check variable name. Is the variable name one of the abbreviations?
    argument_check.check_requested_variable(variable)
    # is it a sfc or pressure level field. # output sfc or level type
    level_type = output_sfc_or_plev(variable)

    # if level_type == plevs and plevs=None, output_plevs. Will only give troposphere for q. 
    # work out appropriate pressure levels
    if level_type == 'pressure':
        if plevs is None:
            plevs = output_plevs(variable)
        else:
            print (f"Downloading the requested pressure levels: {plevs}") # if not, use request plevs.
        # check plevs
        argument_check.check_plevs(plevs,variable)
    else:
        print (f"Downloading the following level type: {level_type}")
        plevs=None

    # get ECDS version of variable name. - WILL WRITE UP IN OCTOBER 2025!
    #ecds_varname = variable_output.output_ECDS_variable_name(variable)
    ecds_varname=None

    # get webapi param
    webapi_param = output_webapi_variable_name(variable) # temporary until move to ECDS (Aug - Oct).

    # check model is in acceptance list and get origin code!
    argument_check.check_model_name(model,fcdate)
    # get origin id
    origin_id = output_originID(model,fcdate)

    # get fc_enslags
    # get lagged ensemble details
    if fc_enslags is None:
        fc_enslags = output_fc_lags(origin_id,fcdate)
    # after gathering fc_enslags, check all ensemble lags are negative or zero and whole numbers as they can be user-inputted.
    argument_check.check_fc_enslags(fc_enslags)

    # if leadtime_hour = None, get leadtime_hour (output all hours).
    if leadtime_hour is None:
        leadtime_hour = output_leadtime_hour(variable,origin_id,fcdate,fc_enslags) # the function outputs an array of hours. This is the leadtime used during download.
    else:
        leadtime_hour = np.array(leadtime_hour) # make leadtime hour an array
    print (f"For the following variable '{variable}' using the following leadtimes '{leadtime_hour}'.")

    # check leadtime_hours (as individuals can choose own leadtime_hours).
    argument_check.check_leadtime_hours(leadtime_hour,variable,origin_id,fcdate)

    # check fcdate.
    argument_check.check_fcdate(fcdate,origin_id)

    # check dataformat
    argument_check.check_dataformat(data_format)

    # check area selection
    argument_check.check_area_selection(area)

    return level_type, plevs, webapi_param, ecds_varname, origin_id, leadtime_hour, fc_enslags

def check_and_output_all_hc_arguments(variable,origin_id,fcdate,rfyears=None):
    ''' Function that will output all the necessary arguments to download reforecast data
    '''
    # get the date of the reforecast model
    rf_model_date = get_hindcast_model_date(origin_id,fcdate)

    # get the reforecast years
    if rfyears is None:
        rfyears = get_hindcast_year_span(origin_id,fcdate)
    # after computing reforecast years, check the chosen set
    argument_check.check_requested_reforecast_years(rfyears,origin_id,fcdate)

    return rf_model_date, rfyears


